/**
 TULCS.java
 Created on February 6, 2007, 12:37 PM
 
 
 SequenceMiner: multiple anomaly classifier
 
 @author: Eugene Turkov, {eturkov@email.arc.nasa.gov}
 @version: 1.0
 
 NASA Ames Research Center
 Computational Sciences Division
 Intelligent Data Understanding Group (IDU)
 Moffett Field, CA 94035
 */

package turkov.math.algorithms;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

public class TULCS
{

	static int binarySearch4Locn(int[] row,int low,int high,int key){
		
//		 Method added by Suratna Budalakoti (02-22-2009).
		
		int mid;
		while(high-low>1){
			mid = (int) Math.floor((low+high)/2);
			if(row[mid] > key)
				high = mid;
			else
				low = mid;
	}

	if(row[low]!=key)
		low = -1;
	
	return low;
	}

	
	public static int getLenghOfLCSUsingIndex(Object[] seq1,HashMap<Object,ArrayList<Integer>> index,int len2){

//		 Method added by Suratna Budalakoti (02-22-2009).
		
		int matchesNum,matchLocn,newValue,val2Search,startPoint,endPoint,matchPoint;
		Object symbol;
		int len1 = seq1.length;
		
		int[]row = new int[len2+2];
		
//		for(int i=0;i<=len2+1;i++)
//			row[i]=0;

//		System.out.println("len1= "+len1);
		ArrayList<Integer> idxArray;
		for(int i=0;i<len1;i++)
		{

//			for(int j=0;j<=len2+1;j++)
				//System.out.print(row[j]+" ");
//			System.out.println(" ");

			symbol = seq1[i];

			idxArray = (ArrayList<Integer>) index.get(symbol);
			
			if(idxArray==null)
			{
//				System.out.print(symbol+" ");
				continue;
			}
			matchesNum = idxArray.size();
//			System.out.print(symbol+" "+matchesNum+" ");
			
			
			for(int matchCnt=1;matchCnt<matchesNum;matchCnt++)
			{
				matchLocn = idxArray.get(matchCnt).intValue();
				newValue = row[matchLocn-1] + 1;
				row[matchLocn] = newValue;
				
				if(newValue > row[matchLocn+1])
				{
					val2Search = newValue;
					startPoint = matchLocn + 1;
					endPoint = idxArray.get(matchCnt-1).intValue();

					if(newValue > row[endPoint-1])
						matchPoint = endPoint;
					else{
						matchPoint = TULCS.binarySearch4Locn(row, startPoint, endPoint, val2Search);
	
						if(matchPoint==-1)
							matchPoint = endPoint;
					}

					Arrays.fill(row,matchLocn+1,matchPoint,newValue);
				}
			}
		}
		return row[len2];
		
	}

	
	public static HashMap<Object, ArrayList<Integer>> getSequenceIndex(Object[] seq){
		
		// Method added by Suratna Budalakoti (02-22-2009).
		
		int len = seq.length;
		
		Integer lenIntObj = new Integer(len+2);
		
		Object curSymbol;
		ArrayList<Integer> curArList;

		HashMap<Object, ArrayList<Integer>> index = new HashMap<Object, ArrayList<Integer>>();
	
		for(int i=len-1;i>=0;i--){
			curSymbol = seq[i];
			curArList = index.get(curSymbol);

			if(curArList==null){
				curArList = new ArrayList<Integer>();
				curArList.add(lenIntObj);
				}
			curArList.add(new Integer(i+1));
			index.put(curSymbol, curArList);

		}
		
		return index;
		
	}

	
    public static int getLengthOfLongestCommonSubsequence(Object[] a, Object[] b)
    {
        int n = a.length;
        int m = b.length;
        int current[] = new int[m+1];
        int previous[] = new int[m+1];
        int row, column;
        
        // It is important to use <=, not <.  The next two for-loops are initialization

        
        // This is the main dynamic programming loop that computes the score and
        // backtracking arrays.
        for(row = 1; row <= n; ++row)
        {
            for(column = 1; column <= m; ++column)
            {                
                if( a[row-1].equals(b[column-1]))
                    current[column] = previous[column - 1] + 1;
                
                else
                    current[column] = previous[column-1];
                
                if( previous[column] >= current[column] )
                    current[column] = previous[column];
                
                if( current[column-1] >= current[column] )
                    current[column] = current[column-1];
            }
            
            previous = current;
            current = new int[previous.length];
        }
        
        return previous[m];
    }
    
    public static Object[] getLongestCommonSubsequence(Object[] a, Object[] b)
    {
        int n = a.length;
        int m = b.length;
        int S[][] = new int[n+1][m+1];
        int R[][] = new int[n+1][m+1];
        int ii, jj;
        
        // It is important to use <=, not <.  The next two for-loops are initialization
        for(ii = 0; ii <= n; ++ii)
        {
            S[ii][0] = 0;
            R[ii][0] = UP;
        }
        for(jj = 0; jj <= m; ++jj)
        {
            S[0][jj] = 0;
            R[0][jj] = LEFT;
        }
        
        // This is the main dynamic programming loop that computes the score and
        // backtracking arrays.
        for(ii = 1; ii <= n; ++ii)
        {
            for(jj = 1; jj <= m; ++jj)
            {
                
                if( a[ii-1].equals(b[jj-1]))
                {
                    S[ii][jj] = S[ii-1][jj-1] + 1;
                    R[ii][jj] = UP_AND_LEFT;
                }
                
                else
                {
                    S[ii][jj] = S[ii-1][jj-1] + 0;
                    R[ii][jj] = NEITHER;
                }
                
                if( S[ii-1][jj] >= S[ii][jj] )
                {
                    S[ii][jj] = S[ii-1][jj];
                    R[ii][jj] = UP;
                }
                
                if( S[ii][jj-1] >= S[ii][jj] )
                {
                    S[ii][jj] = S[ii][jj-1];
                    R[ii][jj] = LEFT;
                }
            }
        }
        
        // The length of the longest substring is S[n][m]
        ii = n;
        jj = m;
        int pos = S[ii][jj] - 1;
        Object[] lcs = new Object[ pos+1 ];
        
        // Trace the backtracking matrix.
        while( ii > 0 || jj > 0 )
        {
            if( R[ii][jj] == UP_AND_LEFT )
            {
                ii--;
                jj--;
                lcs[pos--] = a[ii];
            }
            
            else if( R[ii][jj] == UP )
                ii--;
            
            else if( R[ii][jj] == LEFT )
                jj--;
        }
        
        return lcs;
    }

    private static final short NEITHER     = 0;
    private static final short UP          = 1;
    private static final short LEFT        = 2;
    private static final short UP_AND_LEFT = 3;  
}
